import json
import os
import re
from typing import Any, Dict, List
import jsonpointer
from jsonpointer import JsonPointerException
import requests
import copy
from avrotize.common import avro_name, generic_type
from avrotize.dependency_resolver import inline_dependencies_of, sort_messages_by_dependencies
from urllib.parse import ParseResult, urljoin, urlparse


class JsonToAvroConverter:

    def __init__(self) -> None:
        self.imported_types: Dict[Any, Any] = {}
        self.root_namespace = "example.com"

    def is_empty_type(self, avro_type):
        """Check if the Avro type is an empty type."""
        if len(avro_type) == 0:
            return True
        if isinstance(avro_type, list):
            return all(self.is_empty_type(t) for t in avro_type)
        if isinstance(avro_type, dict):
            if not 'type' in avro_type:
                return True
        return False

    def merge_schemas(self, schemas_arg: list, avro_schemas: list, type_name: str | None = None) -> str | list | dict:
        """Merge multiple Avro type schemas into one."""
        merged_schema: Dict[str, list, dict] = {}
        schemas = []
        for schema_entry in schemas_arg:
            if isinstance(schema_entry, list):
                if len(schema_entry) == 0:
                    continue
                remaining_list = copy.deepcopy(schemas_arg)
                remaining_list.remove(schema_entry)
                for schema in schema_entry:
                    current_list = copy.deepcopy(remaining_list)
                    current_list.append(copy.deepcopy(schema) if isinstance(schema, dict) or isinstance(schema,list) else schema)
                    schemas.append(self.merge_schemas(current_list, avro_schemas, type_name))
                break
            elif isinstance(schema_entry, dict):
                if len(schema_entry) == 0:
                    continue
                schemas.append(copy.deepcopy(schema_entry))
            else:
                schemas.append(copy.deepcopy(schema_entry) if isinstance(schema_entry,list) else schema_entry)

        if len(schemas) == 1 and isinstance(schemas[0], str):
            return schemas[0]

        if type_name:
            merged_schema["name"] = type_name
        
        for schema in schemas:
            if 'dependencies' in schema:
                deps: List[str] = merged_schema.get('dependencies', [])
                deps.extend(schema['dependencies'])
                merged_schema['dependencies'] = deps
            if (isinstance(schema, list) or isinstance(schema, dict)) and len(schema) == 0:
                continue
            if isinstance(schema, str):
                sch = next((s for s in avro_schemas if s.get('name') == schema), None)
                if sch:
                    merged_schema.update(sch)
                else:
                    merged_schema['type'] = schema
            elif 'type' not in schema or 'type' not in merged_schema:
                merged_schema.update(schema)
            else:
                if 'type' in merged_schema and schema['type'] != merged_schema['type']:
                    merged_schema['type'] = [schema['type'],merged_schema['type']]
                if not type_name:
                    merged_schema['name'] = avro_name(merged_schema.get('name','') + schema.get('name',''))
                if 'fields' in schema:
                    if 'fields' in merged_schema:
                        for field in schema['fields']:
                            if field not in merged_schema['fields']:
                                merged_schema['fields'].append(field)
                            else:
                                merged_schema_field = next(f for f in merged_schema['fields'] if f.get('name') == field.get('name'))
                                if merged_schema_field["type"] != field["type"]:
                                    merged_schema_field["type"] = [field["type"],merged_schema_field["type"]]
                                if "doc" in field and "doc" not in merged_schema_field:
                                    merged_schema_field["doc"] = field["doc"]
                    else:
                        merged_schema['fields'] = schema['fields']
                                
        return merged_schema


    def ensure_type(self, type: dict | str | list) -> dict | str | list:
        if isinstance(type, str) or isinstance(type, list) or 'type' in type:
            return type
        
        type['type'] = generic_type()
        return type


    def json_schema_primitive_to_avro_type(self, json_primitive: str | list, format: str | None, enum: list | None, field_name: str, dependencies: list) -> str | dict[str,Any] | list:
        """Convert a JSON-schema primitive type to Avro primitive type."""

        if isinstance(json_primitive, list):
            if enum:
                json_primitive = "string" 
            else:
                union = []
                for item in json_primitive:
                    avro_primitive = self.json_schema_primitive_to_avro_type(item, format, enum, field_name, dependencies)
                    union.append(avro_primitive)
                return union

        if json_primitive == 'string':
            avro_primitive = 'string'
        elif json_primitive == 'integer':
            avro_primitive = 'int'
        elif json_primitive == 'number':
            avro_primitive = 'float'
        elif json_primitive == 'boolean':
            avro_primitive = 'boolean'
        else:
            dependencies.append(json_primitive)
            avro_primitive = json_primitive

        if format:
            if format in ('date-time', 'date'):
                avro_primitive = {'type': 'int', 'logicalType': 'date'}
            elif format in ('time'):
                avro_primitive = {'type': 'int', 'logicalType': 'time-millis'}
            elif format in ('duration'):
                avro_primitive = {'type': 'fixed', 'size': 12, 'logicalType': 'duration'}
            elif format in ('uuid'):
                avro_primitive = {'type': 'string', 'logicalType': 'uuid'}
        
        if enum:
            # replace white space with underscore
            enum = [avro_name(e) for e in enum if isinstance(e, str) and e != ""] 
            # purge duplicates
            enum = list(dict.fromkeys(enum))
            if len(enum) > 0:
                avro_primitive = {"type": "enum", "symbols": enum, "name": avro_name(field_name + "_enum")}
            else:
                avro_primitive = "string"
        return avro_primitive


    def fetch_content(self, url: str | ParseResult):
        # Parse the URL to determine the scheme
        if isinstance(url, str):
            parsed_url = urlparse(url)
        else:
            parsed_url = url
        scheme = parsed_url.scheme

        # Handle HTTP and HTTPS URLs
        if scheme in ['http', 'https']:
            try:
                response = requests.get(url)
                response.raise_for_status()  # Raises an HTTPError if the response status code is 4XX/5XX
                return response.text
            except requests.RequestException as e:
                return f'Error fetching {url}: {e}'

        # Handle file URLs
        elif scheme == 'file':
            # Remove the leading 'file://' from the path for compatibility
            file_path = parsed_url.netloc
            if not file_path:
                file_path = parsed_url.path
            # On Windows, a file URL might start with a '/' but it's not part of the actual path
            if os.name == 'nt' and file_path.startswith('/'):
                file_path = file_path[1:]
            try:
               with open(file_path, 'r') as file:
                    return file.read()
            except Exception as e:
                return f'Error reading file at {file_path}: {e}'

        else:
            return f'Unsupported URL scheme: {scheme}'

    def resolve_reference(self, json_type: dict, base_uri: str, json_doc: dict):
        """Resolve a JSON Pointer reference or a JSON $ref reference."""
        try:
            ref = json_type['$ref']
            content = None
            url = urlparse(ref)
            if url.scheme:
                content = self.fetch_content(ref)
            elif url.path:
                file_uri = urljoin(base_uri, url.path)
                content = self.fetch_content(file_uri)
            if content:
                try:
                    json_schema = json.loads(content)
                    # resolve the JSON Pointer reference, if any
                    if url.fragment:
                        json_schema = jsonpointer.resolve_pointer(json_schema, url.fragment)
                    return json_schema
                except json.JSONDecodeError:
                    raise Exception(f'Error decoding JSON from {ref}')
            
            if url.fragment:
                json_pointer = url.fragment
                ref_schema = jsonpointer.resolve_pointer(json_doc, json_pointer)
                if ref_schema:
                    return ref_schema
        except JsonPointerException as e:
            raise Exception(f'Error resolving JSON Pointer reference for {base_uri}')
        return json_type
            
            

    def json_type_to_avro_type(self, json_type: str | dict, record_name: str, field_name: str, namespace : str, dependencies: list, json_schema: dict, base_uri: str, avro_schema: list, record_stack: list) -> dict | list | str:
        """Convert a JSON type to Avro type."""

        avro_type: list | dict | str = {}
        local_name = avro_name(field_name if field_name else record_name)
        qualified_name = namespace + "." + local_name    

        if isinstance(json_type, dict):

            if 'if' in json_type or 'then' in json_type or 'else' in json_type or 'dependentSchemas' in json_type or 'dependentRequired' in json_type:
                print('WARNING: Conditional schema is not supported and will be ignored.')
                if 'if' in json_type:
                    del json_type['if']
                if 'then' in json_type:
                    del json_type['then']
                if 'else' in json_type:
                    del json_type['else']
                if 'dependentSchemas' in json_type:
                    del json_type['dependentSchemas']
                if 'dependentRequired' in json_type:
                    del json_type['dependentRequired']

            base_type = json_type.copy()
            if 'oneOf' in base_type:
                del base_type['oneOf']
            if 'anyOf' in base_type:
                del base_type['anyOf']
            if 'allOf' in base_type:
                del base_type['allOf']
            json_types = []            

            if 'allOf' in json_type:
                merged_type = copy.deepcopy(base_type)
                allof = json_type['allOf']
                for allof_option in allof:
                    merged_type = {**allof_option, **merged_type}
                json_types.append(merged_type)

            if 'oneOf' in json_type:
                oneof = json_type['oneOf']
                if len(json_types) == 0:
                    for oneof_option in oneof:
                        json_types.append({**base_type, **oneof_option})
                else:
                    new_json_types = []
                    for oneof_option in oneof:
                        for json_type_option in json_types:
                            json_type_option = {**json_type_option, **oneof_option}
                            new_json_types.append(json_type_option)
                    json_types = new_json_types
            
            if 'anyOf' in json_type:
                anyof = json_type['anyOf']
                if len(json_types) == 0:
                    for anyof_option in anyof:
                        json_types.append({**base_type, **anyof_option})
                else:
                    new_json_types = []
                    for anyof_option in anyof:
                        for json_type_option in json_types:
                            json_type_option = {**json_type_option, **anyof_option}
                            new_json_types.append(json_type_option)
                    json_types = new_json_types

            if len(json_types) > 0:
                if len(json_types) == 1:
                    avro_type = self.json_type_to_avro_type(json_types[0], record_name, field_name, namespace, dependencies, json_schema, base_uri, avro_schema, record_stack)
                    if isinstance(avro_type, dict) and self.is_empty_type(avro_type):
                        avro_type['type'] = generic_type()
                    return avro_type
                else:
                    subtypes = []
                    count = 1
                    type_deps = []
                    for json_type_option in json_types:
                        # we only set the field_name if this is not a type reference
                        sub_field_name = avro_name(local_name + "_" + str(count)) if not isinstance(json_type_option, dict) or not '$ref' in json_type_option else None
                        avro_subtype = self.json_type_to_avro_type(json_type_option, record_name, sub_field_name, namespace, dependencies, json_schema, base_uri, avro_schema, record_stack)
                        if isinstance(avro_subtype, dict) and 'dependencies' in avro_subtype:
                            type_deps.extend(avro_subtype['dependencies'])
                            del avro_subtype['dependencies']
                        if not self.is_empty_type(avro_subtype):
                            if isinstance(avro_subtype, list):
                                subtypes.extend(avro_subtype)
                            else:
                                subtypes.append(avro_subtype)
                        count += 1
                    if len(subtypes) == 1:
                        return subtypes[0]
                    return subtypes

            if 'properties' in json_type and not 'type' in json_type:
                json_type['type'] = 'object'

            if 'description' in json_type and isinstance(avro_type, dict):
                avro_type['doc'] = json_type['description']

            if 'title' in json_type and isinstance(avro_type, dict):
                avro_type['name'] = avro_name(json_type['title'])

            # first, pull in any referenced definitions and merge with this schema
            if '$ref' in json_type:
                ref = json_type['$ref']
                if ref in self.imported_types:
                    type_ref = self.imported_types[ref]
                    if isinstance(type_ref, str):
                        dependencies.append(type_ref)
                    return type_ref
                else:
                    new_base_uri = urljoin(base_uri, json_type['$ref'])
                    resolved_json_type = self.resolve_reference(json_type, base_uri, json_schema)
                    if self.is_empty_type(json_type): 
                        # it's a standalone reference, so will import the type into the schema 
                        # and reference it like it was in the same file
                        type_name = record_name
                        type_namespace = namespace
                        parsed_ref = urlparse(ref)
                        if parsed_ref.fragment:
                            type_name = avro_name(parsed_ref.fragment.split('/')[-1])
                            sub_namespace = '.'.join(parsed_ref.fragment.split('/')[1:-1])
                            type_namespace = self.root_namespace + '.' + sub_namespace if len(sub_namespace) > 0 else self.root_namespace
                        
                        # registering in imported_types ahead of resolving to prevent circular references
                        self.imported_types[ref] = type_name
                        # resolve type
                        deps: List[str] = []
                        avro_type = self.json_type_to_avro_type(resolved_json_type, type_name, field_name, type_namespace, deps, json_schema, new_base_uri, avro_schema, record_stack)
                        if isinstance(avro_type, list) or (not isinstance(avro_type, dict) or not avro_type.get('type') == "record"):
                            if isinstance(avro_type, dict) and not 'type' in avro_type:
                                print(f"WARNING: no type definition for {ref} in record {record_name}: {json.dumps(avro_type)}")
                                avro_type = generic_type()

                            avro_type = {
                                "type": "record",
                                "name": type_name,
                                "namespace": type_namespace,
                                "fields": [
                                    {
                                        "name": "value",
                                        "type": avro_type
                                    }
                                ]
                            }

                        if isinstance(avro_type, dict) and len(deps) > 0:
                            avro_type['dependencies'] = deps
                        self.imported_types[ref] = copy.deepcopy(avro_type)
                        if isinstance(avro_type, dict) and 'name' in avro_type:
                            existing_type = next((t for t in avro_schema if t.get('name') == avro_type['name'] and t.get('namespace') == avro_type.get('namespace') ), None)
                            if not existing_type:
                                avro_schema.append(avro_type)
                            full_name = avro_type.get('namespace','')+'.'+avro_type['name'] if 'namespace' in avro_type else avro_type['name']
                            self.imported_types[ref] = full_name
                            dependencies.append(full_name)
                            avro_type = full_name                    
                    else:
                        # it's a reference within a definition, so we will turn this into an inline type
                        json_type.update(resolved_json_type)
                        del json_type['$ref']
                        avro_type = self.json_type_to_avro_type(json_type, record_name, field_name, namespace, dependencies, json_schema, new_base_uri, avro_schema, record_stack)
                        if isinstance(avro_type, dict) and 'name' in avro_type:
                            self.imported_types[ref] = avro_type['name']
                        else:
                            self.imported_types[ref] = avro_type
                    
            # if 'const' is present, make this an enum
            if 'const' in json_type:
                const = json_type['const']
                avro_type = self.merge_schemas([avro_type, {"type": "enum", "symbols": [const], "name": avro_name(local_name)}], avro_schema, local_name)

            json_object_type = json_type.get('type')
            if json_object_type:
                if json_object_type == 'array':
                    if 'items' in json_type:
                        avro_type = self.merge_schemas([avro_type, {"type": "array", "items": self.json_type_to_avro_type(json_type['items'], record_name, field_name, namespace, dependencies, json_schema, base_uri, avro_schema, record_stack)}], avro_schema, avro_type.get('name', local_name))
                    else:
                        avro_type = self.merge_schemas([avro_type, {"type": "array", "items": generic_type()}], avro_schema, avro_type.get('name', local_name))
                elif json_object_type == 'object' or 'object' in json_object_type:
                    record_stack.append(record_name)
                    avro_type = self.merge_schemas([avro_type, self.json_schema_object_to_avro_record(local_name, json_type, namespace, json_schema, base_uri, avro_schema, record_stack)], avro_schema, avro_type.get('name', local_name)  if isinstance(avro_type,dict) else local_name)
                    if isinstance(avro_type, dict) and 'dependencies' in avro_type:
                        dependencies.extend(avro_type['dependencies'])
                        del avro_type['dependencies']
                    record_stack.pop()
                else:
                    local_name = field_name = field_name if field_name else local_name+"_value"
                    avro_type = self.json_schema_primitive_to_avro_type(json_object_type, json_type.get('format'), json_type.get('enum'), field_name, dependencies)
            elif 'enum' in json_type:
                local_name = field_name = field_name if field_name else local_name
                enum_namespace = namespace + '.' + record_stack[-1] + field_name 
                avro_type = self.merge_schemas([avro_type,self.json_schema_primitive_to_avro_type("string", json_type.get('format'), json_type.get('enum'), field_name, dependencies)], avro_schema, avro_type.get('name', local_name)  if isinstance(avro_type,dict) else local_name)
                if isinstance(avro_type, dict):
                    avro_type['namespace'] = enum_namespace
        else:
            if isinstance(json_type, dict):
                avro_type = self.merge_schemas([avro_type,self.json_schema_primitive_to_avro_type(json_type, json_type.get('format'), json_type.get('enum'), field_name, dependencies)], avro_schema, avro_type.get('name', local_name) if isinstance(avro_type,dict) else local_name)
            else:
                avro_type = self.merge_schemas([avro_type,self.json_schema_primitive_to_avro_type(json_type, None, None, field_name, dependencies)], avro_schema, avro_type.get('name', local_name) if isinstance(avro_type,dict) else local_name)
        
        if isinstance(avro_type, dict) and 'name' in avro_type:
            if not 'namespace' in avro_type:
                avro_type['namespace'] = namespace
            existing_type = next((t for t in avro_schema if t.get('name') == avro_type['name'] and t.get('namespace') == avro_type.get('namespace') ), None)
            if existing_type:
                existing_type_name = existing_type.get('namespace') + '.' + existing_type.get('name')
                if not existing_type_name in dependencies:
                    dependencies.append(existing_type_name)
                return existing_type_name
            avro_type['name'] = local_name
        
        return avro_type

        

    def json_schema_object_to_avro_record(self, name: str, json_object: dict, namespace: str, json_schema: dict, base_uri: str, avro_schema: list, record_stack: list) -> dict:
        """Convert a JSON schema object declaration to an Avro record."""

        dependencies = []

        # handle top-level allOf, anyOf, oneOf
        if isinstance(json_object, dict) and ('allOf' in json_object or 'oneOf' in json_object or 'anyOf' in json_object):
            type = self.json_type_to_avro_type(json_object, name, "value", namespace, dependencies, json_schema, base_uri, avro_schema, record_stack)
            if isinstance(type, list):
                type = {
                            "type": "record",
                            "name": name,
                            "namespace": namespace,
                            "fields": [
                                {
                                    "name": "value",
                                    "type": type
                                }
                            ]
                        }
            if dependencies and isinstance(type, dict):
                type['dependencies'] = dependencies
            return type    
        
        title = json_object.get('title')
        record_name = avro_name(name if name else title if title else None)
        if record_name == None:
            raise ValueError(f"Cannot determine record name for json_object {json_object}")
        if len(record_stack) > 0:
            namespace = namespace + '.' + record_stack[-1]
        avro_record = {
            'type': 'record', 
            'name': record_name,
            'namespace': namespace,
            'fields': []
        }
        # we need to prevent circular dependencies, so we will maintain a stack of the in-progress 
        # records and will resolve the cycle as we go
        if record_name in record_stack:
            # to break the cycle, we will use a containment type that references 
            # the record that is being defined
            ref_name = avro_name(record_name + "_ref")
            return {
                    "type": "record",
                    "name": ref_name,
                    "namespace": namespace,
                    "fields": [
                        {
                            "name": record_name,
                            "type": record_name
                        }
                    ]
                }
        record_stack.append(record_name)

        required_fields = json_object.get('required', [])
        if 'properties' in json_object:
            for field_name, field in json_object['properties'].items():
                # skip fields with an bad or empty type
                if not isinstance(field, dict):
                    continue
                avro_field_type = self.ensure_type(self.json_type_to_avro_type(field, record_name, field_name, namespace, dependencies, json_schema, base_uri, avro_schema, record_stack))
                if 'dependencies' in avro_field_type:
                    dependencies.extend(avro_field_type['dependencies'])
                    del avro_field_type['dependencies']
                    
                if avro_field_type is None:
                    raise ValueError(f"avro_field_type is None for field {field_name}")
                
                if isinstance(avro_field_type,dict) and 'type' in avro_field_type and not avro_field_type['type'] in ["array", "map", "record", "enum", "fixed"]:
                    avro_field_type = avro_field_type['type']               
                
                if not field_name in required_fields and not 'null' in avro_field_type:
                    if isinstance(avro_field_type, list):
                        avro_field_type.append('null')
                        avro_field = {"name": avro_name(field_name), "type": avro_field_type}
                    else:
                        avro_field = {"name": avro_name(field_name), "type": ["null", avro_field_type]}
                else:
                    avro_field = {"name": avro_name(field_name), "type": avro_field_type}
                
                if field.get('description'):
                    avro_field['doc'] = field['description']

                avro_record["fields"].append(avro_field)
            
            if 'additionalProperties' in json_object and isinstance(json_object['additionalProperties'], dict):
                additional_props = json_object['additionalProperties']
                values_type = self.json_type_to_avro_type(additional_props, record_name, record_name + "_extensions", namespace, dependencies, json_schema, base_uri, avro_schema, record_stack)
                if self.is_empty_type(values_type):
                    values_type = generic_type()
                avro_record['fields'].append(
                    {
                        "name": record_name, 
                        "namespace": namespace,
                        "type": { 
                            "type": "map", 
                            "values": values_type
                            }
                    })
            elif 'patternProperties' in json_object and isinstance(json_object['patternProperties'], dict):
                pattern_props = json_object['patternProperties']
                for pattern_name, props in pattern_props.items():
                    pattern_name = re.sub(r'[^a-zA-Z0-9_]', '_', pattern_name)
                    if pattern_name == "":
                        pattern_name = "extensions"
                    prop_type = self.ensure_type(self.json_type_to_avro_type(props, record_name, pattern_name, namespace, dependencies, json_schema, base_uri, avro_schema, record_stack))
                    if self.is_empty_type(prop_type):
                        raise ValueError(f"prop_type for pattern name {pattern_name} and record_name {record_name} is empty")
                    avro_record['fields'].append(
                        {
                            "name": avro_name(pattern_name), 
                            "namespace": namespace,
                            "type": {
                                "type": "map",
                                "values": copy.deepcopy(prop_type)
                            }
                        })
        else:
            if 'additionalProperties' in json_object and isinstance(json_object['additionalProperties'], dict):
                additional_props = json_object['additionalProperties']
                values_type = self.json_type_to_avro_type(additional_props, record_name, record_name + "_extensions", namespace, dependencies, json_schema, base_uri, avro_schema, record_stack)
                if self.is_empty_type(values_type):
                    values_type = generic_type()
                avro_record['fields'].append(
                    {
                        "name": "values", 
                        "type": { 
                            "type": "map", 
                            "values": copy.deepcopy(values_type)
                        }
                    })
            elif 'patternProperties' in json_object and isinstance(json_object['patternProperties'], dict):
                pattern_props = json_object['patternProperties']
                for pattern_name, props in pattern_props.items():
                    pattern_name = re.sub(r'[^a-zA-Z0-9_]', '_', pattern_name)
                    if pattern_name == "":
                        pattern_name = "extensions"
                    type = self.ensure_type(self.json_type_to_avro_type(props, record_name, pattern_name, namespace, dependencies, json_schema, base_uri, avro_schema, record_stack))
                    avro_record['fields'].append(
                        {
                            "name": avro_name(pattern_name), 
                            "namespace": namespace,
                            "type": {
                                "type": "map",
                                "values": copy.deepcopy(type)
                            }
                        })
            elif 'type' in json_object and (json_object['type'] == 'object' or 'object' in json_object['type']) and \
                not 'allOf' in json_object and not 'oneOf' in json_object and not 'anyOf' in json_object:
                avro_record['fields'].append(
                    {
                        "name": "values",
                        "type": {
                            "type": "map",
                            "values": generic_type()
                        }
                    })
            elif 'type' in json_object and (json_object['type'] == 'array' or 'array' in json_object['type']) and \
                not 'allOf' in json_object and not 'oneOf' in json_object and not 'anyOf' in json_object:
                if 'items' in json_object:
                    avro_type = self.json_type_to_avro_type(json_object['items'], record_name, 'values', namespace, dependencies, json_schema, base_uri, avro_schema, record_stack)
                else:
                    avro_type = generic_type()
                avro_record['fields'].append(
                    {
                        "name": "values",
                        "type": {
                            "type": "array", 
                            "items": copy.deepcopy(avro_type)
                            }
                    })
            else:
                avro_record = {
                    "name": record_name,
                    "namespace": namespace,
                }
        
        if 'description' in json_object:
            avro_record['doc'] = json_object['description']
        if len(dependencies) > 0:
            # dedupe the list
            dependencies = list(set(dependencies))    
            avro_record['dependencies'] = dependencies        

        record_stack.pop()
        return avro_record
        

    def jsons_to_avro(self, json_schema: dict | list, namespace: str, base_uri: str) -> list:
        """Convert a JSON-schema to an Avro-schema."""
        avro_schema = []
        record_stack = []

        parsed_url = urlparse(base_uri)
        schema_name = 'record'
        if 'type' in json_schema or 'allOf' in json_schema or 'oneOf' in json_schema or 'anyOf' in json_schema or 'properties' in json_schema:
            self.process_definition(json_schema, namespace, base_uri, avro_schema, record_stack, schema_name, json_schema)
        elif 'swagger' in json_schema or ('definitions' in json_schema and not 'type' in json_schema):
            json_schema_defs = json_schema.get('definitions', {})
            if not json_schema_defs:
                raise ValueError('No definitions found in swagger file')
            for schema_name, schema in json_schema_defs.items():
                if 'type' in schema or 'allOf' in schema or 'oneOf' in schema or 'anyOf' in schema or 'properties' in schema:
                    self.process_definition(json_schema, namespace, base_uri, avro_schema, record_stack, schema_name, schema)
                else:
                    self.process_definition_list(json_schema, namespace, base_uri, avro_schema, record_stack, schema_name, schema.copy())
        elif isinstance(json_schema, list):
            self.process_definition_list(json_schema, namespace, base_uri, avro_schema, record_stack, schema_name, json_schema)
        else:
            raise ValueError('No schema found in input file')
        
        avro_schema = sort_messages_by_dependencies(avro_schema)
        
        if parsed_url.fragment:
            self.imported_types.clear()
            fragment_schema = []
            json_pointer = parsed_url.fragment
            schema_name = parsed_url.fragment.split('/')[-1]
            schema = jsonpointer.resolve_pointer(json_schema, json_pointer)
            avro_schema_item = self.json_schema_object_to_avro_record(schema_name, schema, namespace, json_schema, base_uri, fragment_schema, record_stack)
            inline_dependencies_of(avro_schema, avro_schema_item)
            return avro_schema_item
        
        return avro_schema

    def process_definition_list(self, json_schema, namespace, base_uri, avro_schema, record_stack, schema_name, json_schema_list):
        for sub_schema_name, schema in json_schema_list.items():
            if not isinstance(schema, dict) and not isinstance(schema, list):
                continue
            if 'type' in schema or 'allOf' in schema or 'oneOf' in schema or 'anyOf' in schema or 'properties' in schema:
                self.process_definition(json_schema, namespace, base_uri, avro_schema, record_stack, sub_schema_name, schema)
                continue
            self.process_definition_list(json_schema, namespace, base_uri, avro_schema, record_stack, schema_name, schema)

    def process_definition(self, json_schema, namespace, base_uri, avro_schema, record_stack, schema_name, schema):
        """ Process a schema definition. """
        avro_schema_item_list = self.json_schema_object_to_avro_record(schema_name, schema, namespace, json_schema, base_uri, avro_schema, record_stack)
        if not isinstance(avro_schema_item_list, list) and not isinstance(avro_schema_item_list, dict):
            return
        # the call above usually returns a single record, but we pretend it's normally a list to handle allOf/anyOf/oneOf cases
        if not isinstance(avro_schema_item_list, list):
            avro_schema_item_list = [avro_schema_item_list]
        for avro_schema_item in avro_schema_item_list:
            avro_schema_item['name'] = avro_name(schema_name)
            existing_type = next((t for t in avro_schema if t.get('name') == avro_schema_item['name'] and t.get('namespace') == avro_schema_item.get('namespace') ), None)
            if not existing_type:
                avro_schema.append(avro_schema_item)
        

    def id_to_avro_namespace(self, id: str) -> str:
        """Convert a XSD namespace to Avro Namespace."""
        parsed_url = urlparse(id)
        # strip the file extension 
        path = parsed_url.path.rsplit('.')[0]
        path_segments = path.strip('/').replace('-', '_').split('/')
        reversed_path_segments = reversed(path_segments)
        namespace_suffix = '.'.join(reversed_path_segments)
        namespace_prefix = '.'.join(reversed(parsed_url.hostname.split('.')))
        namespace = f"{namespace_prefix}.{namespace_suffix}"
        return namespace

    def convert_jsons_to_avro(self, json_schema_file_path: str, avro_schema_path: str, namespace: str = None) -> list:
        """Convert JSON schema file to Avro schema file."""
        # turn the file path into a file URI if it's not a URI already
        parsed_url = urlparse(json_schema_file_path)
        if not parsed_url.hostname and not parsed_url.scheme == "file":
            json_schema_file_path = 'file://' + json_schema_file_path
            parsed_url = urlparse(json_schema_file_path)
        content = self.fetch_content(parsed_url.geturl())
        json_schema = json.loads(content)

        if not namespace:
            namespace = parsed_url.geturl().replace('\\','/').replace('-','_').split('/')[-1].split('.')[0]
            # get the $id if present 
            if '$id' in json_schema:
                namespace = self.id_to_avro_namespace(json_schema['$id'])
        self.root_namespace = namespace
        
        # drop the file name from the parsed URL to get the base URI
        avro_schema = self.jsons_to_avro(json_schema, namespace, parsed_url.geturl())
        if len(avro_schema) == 1:
            avro_schema = avro_schema[0]

        # create the directory for the Avro schema file if it doesn't exist
        dir = os.path.dirname(avro_schema_path)
        if dir != '' and not os.path.exists(dir):
            os.makedirs(dir, exist_ok=True)
        with open(avro_schema_path, 'w') as avro_file:
            json.dump(avro_schema, avro_file, indent=4)
        return avro_schema
    

def convert_jsons_to_avro(json_schema_file_path: str, avro_schema_path: str, namespace: str = None) -> list:
    """Convert JSON schema file to Avro schema file."""
    converter = JsonToAvroConverter()
    return converter.convert_jsons_to_avro(json_schema_file_path, avro_schema_path, namespace)


import json
import sys

def convert_avro_to_kusto(avro_schema_path, avro_record_type, kusto_file_path, emit_cloud_events_columns=False):
    schema_file = avro_schema_path
    if not schema_file:
        print("Please specify the avro schema file")
        sys.exit(1)
    with open(schema_file, "r") as f:
        schema_json = f.read()

    # Parse the schema as a JSON object
    schema = json.loads(schema_json)

    if isinstance(schema, list) and avro_record_type:
        schema = next((x for x in schema if x["name"] == avro_record_type), None)
        if schema is None:
            print(f"No record type {avro_record_type} found in the Avro schema")
            sys.exit(1)
    elif not isinstance(schema, dict):
        print("Expected a single Avro schema as a JSON object, or a list of schema records")
        sys.exit(1)

    # Get the name and fields of the top-level record
    table_name = schema["name"]
    fields = schema["fields"]

    # Create a StringBuilder to store the kusto statements
    kusto = []

    # Append the create table statement with the column names and types
    kusto.append(f".create table [{table_name}] (")
    for field in fields:
        column_name = field["name"]
        column_type = convert_avro_type_to_kusto_type(field["type"])
        kusto.append(f"    [{column_name}]: {column_type},")
    if emit_cloud_events_columns:
        kusto.append("    [__type]: string,")
        kusto.append("    [__source]: string,")
        kusto.append("    [__id]: string,")
        kusto.append("    [__time]: datetime,")
        kusto.append("    [__subject]: string,")
    kusto.append(");")
    kusto.append("")

    # Add the doc string as table metadata
    if "doc" in schema:
        doc_string = schema["doc"]
        kusto.append(f".alter table [{table_name}] docstring '{doc_string}';")
        kusto.append("")

    doc_string_statement = []
    for field in fields:
        column_name = field["name"]
        if "doc" in field:
            doc = field["doc"]
            doc_string_statement.append(f"   [{column_name}]: '{doc}',")
    if doc_string_statement and emit_cloud_events_columns:
        doc_string_statement.extend([
            "  [__type] : 'Event type',",
            "  [__source]: 'Context origin/source of the event',",
            "  [__id]: 'Event identifier',",
            "  [__time]: 'Event generation time',",
            "  [__subject]: 'Context subject of the event',"
        ])
    if doc_string_statement:
        kusto.append(".alter table [{table_name}] column-docstrings (")
        kusto.extend(doc_string_statement)
        kusto.append(");")
        kusto.append("")

    with open(kusto_file_path, "w") as kusto_file:
        kusto_file.write("\n".join(kusto))

def convert_avro_type_to_kusto_type(avroType):
    if isinstance(avroType, list):
        # If the type is an array, then it is a union type. Look whether it's a pair of a scalar type and null:
        itemCount = len(avroType)
        if itemCount > 2:
            return "dynamic"
        if itemCount == 1:
            return convert_avro_type_to_kusto_type(avroType[0])
        else:
            first = avroType[0]
            second = avroType[1]
            if isinstance(first, str) and first == "null":
                return convert_avro_type_to_kusto_type(second)
            elif isinstance(second, str) and second == "null":
                return convert_avro_type_to_kusto_type(first)
            else:
                return "dynamic"
    elif isinstance(avroType, dict):
        type = avroType.get("type")
        if type == "enum":
            return "string"
        elif type == "fixed":
            return "dynamic"
        elif type == "string":
            logicalType = avroType.get("logicalType")
            if logicalType == "uuid":
                return "guid"
            return "string"
        elif type == "bytes":
            logicalType = avroType.get("logicalType")
            if logicalType == "decimal":
                return "decimal"
            return "dynamic"
        elif type == "long":
            logicalType = avroType.get("logicalType")
            if logicalType in ["timestamp-millis", "timestamp-micros"]:
                return "datetime"
            if logicalType in ["time-millis", "time-micros"]:
                return "timespan"
            return "long"
        elif type == "int":
            logicalType = avroType.get("logicalType")
            if logicalType == "date":
                return "datetime"
            return "int"
        else:
            return map_scalar_type(type)
    elif isinstance(avroType, str):
        return map_scalar_type(avroType)
    return "dynamic"

def map_scalar_type(type):
    if type == "null":
        return "dynamic"
    elif type == "int":
        return "int"
    elif type == "long":
        return "long"
    elif type == "float":
        return "real"
    elif type == "double":
        return "real"
    elif type == "boolean":
        return "bool"
    elif type == "bytes":
        return "dynamic"
    elif type == "string":
        return "string"
    else:
        return "dynamic"


import re

def avro_name(name):
    """Convert a name into an Avro name."""
    val = re.sub(r'[^a-zA-Z0-9_]', '_', name)
    if re.match(r'^[0-9]', val):
        val = '_' + val
    return val

def generic_type() -> list[str | dict]:
    simple_type_union: list[str | dict] = ["null", "boolean", "int", "long", "float", "double", "bytes", "string"]
    l2 = simple_type_union.copy()
    l2.extend([
        {
            "type": "array",
            "items": simple_type_union
        },
        {
            "type": "map",
            "values": simple_type_union
        }])
    l1 = simple_type_union.copy()
    l1.extend([
        {
            "type": "array",
            "items": l2
        },
        {
            "type": "map",
            "values": l2
        }])
    return l1


import argparse
import json
import os

from avrotize.dependency_resolver import sort_messages_by_dependencies

from . import proto2parser
from . import proto3parser
import re
    
isomorphic_types = ['float', 'double', 'bytes', 'string']
imported_types = {}

def proto_type_to_avro_primitive(proto_type):
    """Map Protobuf types to Avro primitive types."""
    mapping = {
        'google.protobuf.Empty': 'null',  # Special handling may be required
        'bool': 'boolean',
        'int32': 'int',
        'uint32': 'int',
        'sint32': 'int',
        'int64': 'long',
        'uint64': 'long',
        'sint64': 'long',
        'fixed32': 'int',
        'fixed64': 'long',
        'sfixed32': 'int',
        'sfixed64': 'long',
        #'NullValue': 'null', 
        'google.protobuf.Timestamp' : {
            "type": "long",
            "logicalType": "timestamp-micros"
        }
    }
    if proto_type in isomorphic_types:
        return proto_type
    return mapping.get(proto_type, proto_type)  


def convert_proto_to_avro(proto_file_path: str, avro_schema_path: str):
    """Convert Protobuf .proto file to Avro schema."""

    def convert_proto_to_avro_schema(proto_file_path: str) -> list:
        """Convert .proto file to Avro schema."""
        with open(proto_file_path, 'r') as proto_file:
            proto_schema = proto_file.read()

        # determine whether we have proto3 or proto2 and parse the data
        if re.search(r'syntax\s*=\s*"proto3"', proto_schema):
            data = proto3parser.parse(proto_schema)
        else:
            data = proto2parser.parse(proto_schema)

        # get the namespace
        namespace = data.package.value
        # Avro schema header
        avro_schema = []
        
        for import_ in data.imports:
            # handle protobuf imports
            if import_.startswith('google/protobuf/'):
                script_path = os.path.dirname(os.path.abspath(__file__))
                avsc_dir = os.path.join(script_path, 'prototypes')            

                # load the corresponding avsc file from ./prototypes at this script's path into avro_schema
                avsc = f'{avsc_dir}/{import_.replace("google/protobuf/", "").replace(".proto", ".avsc")}'
                with open(avsc, 'r') as avsc_file:
                    types = json.load(avsc_file)
                    for t in types:
                        imported_types[t["namespace"]+"."+t["name"]] = t                
            else:
                # find the path relative to the current directory
                cwd = os.path.join(os.getcwd(),os.path.dirname(proto_file_path))
                import_path = os.path.join(cwd, import_)
                # raise an exception if the imported file does not exist
                if not os.path.exists(import_path):
                    raise FileNotFoundError(f'Import file {import_path} does not exist.')
                                
                avro_schema.extend(convert_proto_to_avro_schema(import_path))
                    
        
        ## Convert message fields
        for _, m in data.messages.items():
            handle_message(m, avro_schema, namespace)        

        ## Convert enum fields
        for _, enum_type in data.enums.items():
            handle_enum(enum_type, avro_schema, namespace)

        # Sort the messages in avro_schema by dependencies
        return sort_messages_by_dependencies(avro_schema)
        
    
    avro_schema = convert_proto_to_avro_schema(proto_file_path)
    ## Convert the Avro schema to JSON and write it to the file
    with open(avro_schema_path, 'w') as avro_file:
        avro_file.write(json.dumps(avro_schema, indent=2))

    print(f'Converted {proto_file_path} to {avro_schema_path}')


def handle_enum(enum_type, avro_schema, namespace):
    """Convert enum fields to avro schema."""
    comment = enum_type.comment.content if enum_type.comment and enum_type.comment.content else None
    if comment:
        # strip slashes, newlines, linefeeds from the comment and then extra whitespace
        comment = comment.replace('//', '').replace('\n','').lstrip().rstrip()

    # create avro schema
    avro_enum = {
            'name': enum_type.name,
            'type': 'enum',
            'namespace' : namespace,
            'symbols': [],
            'dependencies': []
        }

    if comment:
        avro_enum['doc'] = comment

    for value in enum_type.fields:
        avro_enum['symbols'].append(value.name)

    avro_schema.append(avro_enum)


def handle_message(m, avro_schema, namespace):
    """Convert protobuf messages to avro records."""
    dependencies = []

    comment = m.comment.content if m.comment and m.comment.content else None
    if comment:
           # strip slashes, newlines, linefeeds from the comment and then extra whitespace
       comment = comment.replace('//', '').replace('\n','').lstrip().rstrip()
    avro_record = {
            'type': 'record',
            'name': m.name,
            'namespace' : namespace,
            'fields': []
        }
    if comment:
        avro_record['doc'] = comment

        
    for f in m.fields:
        avro_type = get_avro_type_for_field(m, namespace, dependencies, f)
        comment = f.comment.content if f.comment and f.comment.content else None
        if comment:
                # strip slashes, newlines, linefeeds from the comment and then extra whitespace
            comment = comment.replace('//', '').replace('\n','').lstrip().rstrip()

        avro_field = {
                'name': f.name,
                'type': avro_type,
                }
        if comment:
            avro_field['doc'] = comment

        avro_record['fields'].append(avro_field)

    for f in m.oneofs:
        avro_oneof = {
                'name': f.name,
                'type': []
                }
        comment = f.comment.content if f.comment and f.comment.content else None
        if comment:
                # strip slashes, newlines, linefeeds from the comment and then extra whitespace
            comment = comment.replace('//', '').replace('\n','').lstrip().rstrip()
            avro_oneof['doc'] = comment

        for o in f.fields:
            avro_type = get_avro_type_for_field(m, namespace, dependencies, o)
            comment = o.comment.content if o.comment and o.comment.content else None
            if comment:
                # strip slashes, newlines, linefeeds from the comment and then extra whitespace
                comment = comment.replace('//', '').replace('\n','').lstrip().rstrip()
            avro_oneof['type'].append(avro_type)            

        avro_record['fields'].append(avro_oneof)

    if dependencies:
       avro_record['dependencies'] = dependencies    
    avro_schema.append(avro_record)

    for _, mi in m.messages.items():
        handle_message(mi, avro_schema, namespace)        

    ## Convert enum fields
    for _, enum_type in m.enums.items():
        handle_enum(enum_type, avro_schema, namespace)

def get_avro_type_for_field(m, namespace, dependencies, f):
    field_type = None
    is_custom = False
    if f.label == 'repeated' or f.type == 'map':
        field_type = proto_type_to_avro_primitive(f.val_type)
        is_custom = field_type == f.val_type and field_type not in isomorphic_types
    else:
        field_type = proto_type_to_avro_primitive(f.type)
        is_custom = field_type == f.type and field_type not in isomorphic_types

    if is_custom:
        if f.type in imported_types:
            field_type = imported_types[f.type]
            imported_types[f.type] = f.type
        else:
            found = False
            for k, mi in m.messages.items():
                if mi.name == field_type:
                    schema = []
                    handle_message(mi, schema, namespace)
                    del m.messages[k]
                    field_type = schema[0]
                    if 'dependencies' in field_type:
                        dependencies.extend(field_type['dependencies'])
                        del field_type['dependencies']

                    found = True
                    break 
            if not found:
                dependencies.append(field_type)

    if f.label == 'optional':
        field_type = ["null", field_type]
    if f.label == 'repeated':
        avro_type = {
                "type": "array",
                "items" : field_type                    
                }
    elif f.type == 'map':
        avro_type = {
                "type": "map",
                "values" : field_type,
                }
    else:
        avro_type = field_type
    return avro_type





import argparse
from avrotize.asn1toavro import convert_asn1_to_avro
from avrotize.avrotokusto import convert_avro_to_kusto
from avrotize.avrotoparquet import convert_avro_to_parquet

from avrotize.avrotoproto import convert_avro_to_proto
from avrotize.avrototsql import convert_avro_to_tsql
from avrotize.jsonstoavro import convert_jsons_to_avro
from avrotize.prototoavro import convert_proto_to_avro
from avrotize.xsdtoavro import convert_xsd_to_avro


def main():
    parser = argparse.ArgumentParser(description='Convert Proto schema to Avro Schema')

    subparsers = parser.add_subparsers(dest='command')
    a2p_parser = subparsers.add_parser('p2a', help='Convert Avro schema to Proto schema')
    a2p_parser.add_argument('--proto', type=str, help='Path of the proto file', required=True)
    a2p_parser.add_argument('--avsc', type=str, help='Path to the Avro schema file', required=True)

    p2a_parser = subparsers.add_parser('a2p', help='Convert Proto schema to Avro schema')
    p2a_parser.add_argument('--proto', type=str, help='Path to the Proto file', required=True)
    p2a_parser.add_argument('--avsc', type=str, help='Path to the Avro schema file', required=True)
    p2a_parser.add_argument('--naming', type=str, help='Type naming convention', choices=['snake', 'camel', 'pascal'], required=False, default='pascal')
    p2a_parser.add_argument('--allow-optional', action='store_true', help='Enable support for "optional" fields', default=False)

    j2a_parser = subparsers.add_parser('j2a', help='Convert JSON schema to Avro schema')
    j2a_parser.add_argument('--jsons', type=str, help='Path to the JSON schema file', required=True)
    j2a_parser.add_argument('--avsc', type=str, help='Path to the Avro schema file', required=True)
    j2a_parser.add_argument('--namespace', type=str, help='Namespace for the Avro schema', required=False)

    x2a_parser = subparsers.add_parser('x2a', help='Convert XSD schema to Avro schema')
    x2a_parser.add_argument('--xsd', type=str, help='Path to the XSD schema file', required=True)
    x2a_parser.add_argument('--avsc', type=str, help='Path to the Avro schema file', required=True)
    x2a_parser.add_argument('--namespace', type=str, help='Namespace for the Avro schema', required=False)

    a2k_parser = subparsers.add_parser('a2k', help='Convert Avro schema to Kusto schema')
    a2k_parser.add_argument('--avsc', type=str, help='Path to the Avro schema file', required=True)
    a2k_parser.add_argument('--kusto', type=str, help='Path to the Kusto table', required=True)
    a2k_parser.add_argument('--record_type', type=str, help='Record type in the Avro schema', required=False)
    a2k_parser.add_argument('--emit_cloudevents_columns', action='store_true', help='Add CloudEvents columns to the Kusto table', default=False)

    a2tsql_parser = subparsers.add_parser('a2tsql', help='Convert Avro schema to T-SQL schema')
    a2tsql_parser.add_argument('--avsc', type=str, help='Path to the Avro schema file', required=True)
    a2tsql_parser.add_argument('--tsql', type=str, help='Path to the T-SQL table', required=True)
    a2tsql_parser.add_argument('--record_type', type=str, help='Record type in the Avro schema', required=False)
    a2tsql_parser.add_argument('--emit_cloudevents_columns', action='store_true', help='Add CloudEvents columns to the T-SQL table', default=False)

    a2pq_parser = subparsers.add_parser('a2pq', help='Convert Avro schema to Parquet schema')
    a2pq_parser.add_argument('--avsc', type=str, help='Path to the Avro schema file', required=True)
    a2pq_parser.add_argument('--parquet', type=str, help='Path to the Parquet file', required=True)
    a2pq_parser.add_argument('--record_type', type=str, help='Record type in the Avro schema', required=False)
    a2pq_parser.add_argument('--emit_cloudevents_columns', action='store_true', help='Add CloudEvents columns to the Parquet file', default=False)

    asn2a_parser = subparsers.add_parser('asn2a', help='Convert ASN.1 schema to Avro schema')
    asn2a_parser.add_argument('--asn', type=str, nargs='+', help='Path(s) to the ASN.1 schema file(s)', required=True)
    asn2a_parser.add_argument('--avsc', type=str, help='Path to the Avro schema file', required=True)



    args = parser.parse_args()
    if args.command is None:
        parser.print_help()
    elif args.command == 'p2a':
        avro_schema_path = args.avsc
        proto_file_path = args.proto
        convert_proto_to_avro(proto_file_path, avro_schema_path)
    elif args.command == 'a2p':
        proto_schema_path = args.proto
        avro_file_path = args.avsc
        naming = args.naming
        allow_optional = args.allow_optional
        convert_avro_to_proto(avro_file_path, proto_schema_path, naming_mode=naming, allow_optional=allow_optional)
    elif args.command == 'j2a':
        json_schema_file_path = args.jsons
        avro_schema_path = args.avsc
        namespace = args.namespace
        convert_jsons_to_avro(json_schema_file_path, avro_schema_path, namespace=namespace)
    elif args.command == 'x2a':
        xsd_schema_file_path = args.xsd
        avro_schema_path = args.avsc
        namespace = args.namespace
        convert_xsd_to_avro(xsd_schema_file_path, avro_schema_path, namespace=namespace)
    elif args.command == 'a2k':
        avro_schema_path = args.avsc
        avro_record_type = args.record_type
        kusto_file_path = args.kusto
        emit_cloud_events_columns = args.emit_cloudevents_columns
        convert_avro_to_kusto(avro_schema_path, avro_record_type, kusto_file_path, emit_cloud_events_columns)
    elif args.command == 'a2tsql':
        avro_schema_path = args.avsc
        avro_record_type = args.record_type
        tsql_file_path = args.tsql
        emit_cloud_events_columns = args.emit_cloudevents_columns
        convert_avro_to_tsql(avro_schema_path, avro_record_type, tsql_file_path, emit_cloud_events_columns)
    elif args.command == 'a2pq':
        avro_schema_path = args.avsc
        avro_record_type = args.record_type
        parquet_file_path = args.parquet
        emit_cloud_events_columns = args.emit_cloudevents_columns
        convert_avro_to_parquet(avro_schema_path, avro_record_type, parquet_file_path, emit_cloud_events_columns)  
    elif args.command == 'asn2a':
        asn_schema_file_list = args.asn
        if len(asn_schema_file_list) == 1:
            asn_schema_file_list = asn_schema_file_list[0].split(',')
        avro_schema_path = args.avsc
        convert_asn1_to_avro(asn_schema_file_list, avro_schema_path)
    

if __name__ == "__main__":
    try:
        main()
    except Exception as e:
        print("Error: ", str(e))
        exit(1)



# sort the dependencies

import copy


def inline_dependencies_of(avro_schema, record):
    for dependency in record.get('dependencies', []):
        dependency_type = next((x for x in avro_schema if x['name'] == dependency or x.get('namespace','')+'.'+x['name'] == dependency), None)
        if not dependency_type:
            continue
        deps = record.get('dependencies', [])
        for field in record['fields']:                        
            if record['namespace']+'.'+record['name'] != dependency:
                swap_dependency_type(avro_schema, field, dependency, dependency_type, deps)
    del record['dependencies']

    

def sort_messages_by_dependencies(avro_schema):
    """ 
        Sort the messages in avro_schema by their dependencies. Avro Schema requires
        that type definitions must be defined before they are used. This method
        ensures this. Types that have dependencies will be moved at the end of the list.
        If necessary, it will also resolve circular dependencies by inlining the 
        dependent record.

        The method expects all types with dependencies to have a 'dependencies' key in their
        dict that contains a list of types that they depend on.

        Args:
            avro_schema: List of Avro schema records.
    """

    # if all are just strings, then it is already sorted
    if all(isinstance(record, str) for record in avro_schema):
        return avro_schema

    sorted_messages = []
    while avro_schema:
        found = False
        for record in avro_schema:
            if not isinstance(record, dict):
                sorted_messages.append(record)
                avro_schema.remove(record)
                continue
            
            # if this record is not a dependency of any other record, it can be safely emitted now
            #if not any(record.get('namespace','')+'.'+record.get('name') in other_record.get('dependencies', []) for other_record in [x for x in avro_schema if isinstance(x, dict) and 'name' in x]):
            remaining_deps = [dep for dep in record['dependencies'] if not dep in [x.get('namespace','')+'.'+x.get('name','') for x in sorted_messages]] if 'dependencies' in record else []
            if len(remaining_deps) == 0:
                if 'dependencies' in record:
                    del record['dependencies']
                sorted_messages.append(record)
                avro_schema.remove(record)
                found = True
                
        # If there are no records without dependencies, we will grab the first 
        # record with dependencies and start resolving circular dependencies
        if len(avro_schema) > 0 and not found:
            found = False
            for record in avro_schema:
                if isinstance(record, dict) and 'dependencies' in record:
                    remaining_deps = [dep for dep in record['dependencies'] if not dep in [x.get('namespace','')+'.'+x.get('name','') for x in sorted_messages]]
                    if len(remaining_deps) > 0:
                        swap_record_dependencies(avro_schema, record)
                        if isinstance(record, dict) and not 'dependencies' in record:
                            found = True
                        else:
                            remaining_remaining_deps = [dep for dep in record['dependencies'] if not dep in [x.get('namespace')+'.'+x.get('name') for x in sorted_messages]]
                            found = len(remaining_deps) != len(remaining_remaining_deps)
                            if found:
                                break

            if not found:
                found = False
                for record in avro_schema:
                    if isinstance(record, dict) and 'dependencies' in record:            
                        found = True
                        record_deps = copy.deepcopy(record.get('dependencies', []))
                        inline_dependencies_of(avro_schema, record)
                        # fix the dependencies of all records that have this record as a dependency
                        for remaining_schema in avro_schema:
                            if isinstance(remaining_schema, dict) and 'dependencies' in remaining_schema and any(dep in record_deps for dep in remaining_schema['dependencies']):
                                remaining_schema['dependencies'] = [dep for dep in remaining_schema['dependencies'] if not dep in record_deps]
                                qname = record['namespace']+'.'+record['name']
                                if not qname in remaining_schema['dependencies']:
                                    remaining_schema['dependencies'].append(qname)
                        break

                if not found:
                    print('WARNING: There are circular dependencies in the schema, unable to resolve them: {}'.format([x['name'] for x in avro_schema if isinstance(x, dict) and 'dependencies' in x]))
                            
        
    #sorted_messages.reverse()
    return sorted_messages

def swap_record_dependencies(avro_schema, record):
    while 'dependencies' in record and len(record['dependencies']) > 0:
        for dependency in record.get('dependencies', []):
            dependency_type = next((x for x in avro_schema if x['name'] == dependency or x.get('namespace','')+'.'+x['name'] == dependency), None)
            if not dependency_type and dependency in record['dependencies']:
                record['dependencies'].remove(dependency)
            deps = record.get('dependencies', [])
            for field in record['fields']:                        
                if record['name'] != dependency and (record.get('namespace','')+'.'+record['name']) != dependency:
                    swap_dependency_type(avro_schema, field, dependency, dependency_type, deps)
            record['dependencies'] = [dep for dep in deps if dep != record['name'] and record.get('namespace','')+'.'+record['name'] != dep]
            if len(record['dependencies']) == 0:
                del record['dependencies']


def swap_dependency_type(avro_schema, field, dependency, dependency_type, dependencies):
    """ to break circular dependencies, we will inline the dependent record and remove the dependency """
    if not dependency in dependencies:
        return
    if not dependency_type in avro_schema:
        return
    
    # Replace the dependency type with the dependency_type in avro_schema.
    if field['type'] == dependency:
        field['type'] = dependency_type
        if dependency_type in avro_schema:
            avro_schema.remove(dependency_type)
        dependencies.remove(dependency)
        dependencies.extend(dependency_type.get('dependencies', []))
        if 'dependencies' in dependency_type:
            swap_record_dependencies(avro_schema, dependency_type)
           
    # type is a Union?
    elif isinstance(field['type'], list):
        for field_type in field['type']:
            if field_type == dependency:
                field['type'].remove(field_type)
                field['type'].append(dependency_type)
                if dependency_type in avro_schema:
                    avro_schema.remove(dependency_type)
                dependencies.remove(dependency)
                dependencies.extend(dependency_type.get('dependencies', []))
                if 'dependencies' in dependency_type:
                    swap_record_dependencies(avro_schema, dependency_type)
            # type is an object?
            elif isinstance(field_type, dict) and 'type' in field_type and field_type.get('type') == dependency or \
                'items' in field_type and field_type.get('items') == dependency or \
                'values' in field_type and field_type.get('values') == dependency:
                swap_dependency_type(avro_schema, field_type, dependency, dependency_type, dependencies)
            elif isinstance(field_type, list):
                for item in field_type:
                    if item == dependency or isinstance(item, dict) and item.get('type') == dependency:
                        swap_dependency_type(avro_schema, field_type, dependency, dependency_type, dependencies)
            elif isinstance(field_type, dict) and 'type' in field_type:
                swap_dependency_type(avro_schema, field_type, dependency, dependency_type, dependencies)
    elif isinstance(field['type'], dict) and 'type' in field['type']:
        swap_dependency_type(avro_schema, field['type'], dependency, dependency_type, dependencies)
    elif field['type'] == 'array':
        if isinstance(field['items'], list):
            for item in field['items']:
                if item == dependency:
                    field['items'].remove(item)
                    field['items'].append(dependency_type)
                    if dependency_type in avro_schema:
                        avro_schema.remove(dependency_type)
                    dependencies.remove(dependency)
                    dependencies.extend(dependency_type.get('dependencies', []))
                    if 'dependencies' in dependency_type:
                        swap_record_dependencies(avro_schema, dependency_type)
                elif isinstance(item, dict) and 'type' in item and item.get('type') == dependency:
                    swap_dependency_type(avro_schema, item, dependency, dependency_type, dependencies)                
        elif field['items'] == dependency:
            field['items'] = dependency_type
            if dependency_type in avro_schema:
                avro_schema.remove(dependency_type)
            dependencies.remove(dependency)
            dependencies.extend(dependency_type.get('dependencies', []))
            if 'dependencies' in dependency_type:
                swap_record_dependencies(avro_schema, dependency_type)
        elif 'type' in field['items']:
            swap_dependency_type(avro_schema, field['items'], dependency, dependency_type, dependencies)
    elif field['type'] == 'map':
        if isinstance(field['values'], list):
            for item in field['values']:
                if item == dependency:
                    field['values'].remove(item)
                    field['values'].append(dependency_type)
                    if dependency_type in avro_schema:
                        avro_schema.remove(dependency_type)
                    dependencies.remove(dependency)
                    dependencies.extend(dependency_type.get('dependencies', []))
                    if 'dependencies' in dependency_type:
                        swap_record_dependencies(avro_schema, dependency_type)
                elif isinstance(item, dict) and 'type' in item and item.get('type') == dependency:
                    swap_dependency_type(avro_schema, item, dependency, dependency_type, dependencies)
        if field['values'] == dependency:
            field['values'] = dependency_type
            if dependency_type in avro_schema:
                avro_schema.remove(dependency_type)                    
            dependencies.remove(dependency)
            dependencies.extend(dependency_type.get('dependencies', []))
            if 'dependencies' in dependency_type:
                swap_record_dependencies(avro_schema, dependency_type)
        elif 'type' in field['values']:
            swap_dependency_type(avro_schema, field['values'], dependency, dependency_type, dependencies)
    elif field['type'] == 'record':
        for dep_field in field['fields']:
            swap_dependency_type(avro_schema, dep_field, dependency, dependency_type, dependencies)



import json
import sys

def convert_avro_to_tsql(avro_schema_path, avro_record_type, tsql_file_path, emit_cloud_events_columns=False):
    schema_file = avro_schema_path
    if not schema_file:
        print("Please specify the avro schema file")
        sys.exit(1)
    with open(schema_file, "r") as f:
        schema_json = f.read()

    # Parse the schema as a JSON object
    schema_list = schema = json.loads(schema_json)

    if isinstance(schema, list) and avro_record_type:
        schema = next((x for x in schema if x["name"] == avro_record_type), None)
        if schema is None:
            print(f"No record type {avro_record_type} found in the Avro schema")
            sys.exit(1)
    elif not isinstance(schema, dict):
        print("Expected a single Avro schema as a JSON object, or a list of schema records")
        sys.exit(1)

    # Get the name and fields of the top-level record
    table_name = schema["name"]
    fields = schema["fields"]

    # Create a list to store the T-SQL statements
    tsql = []

    # Append the create table statement with the column names and types
    tsql.append(f"CREATE TABLE [{table_name}] (")
    for field in fields:
        column_name = field["name"]
        column_type = convert_avro_type_to_tsql_type(field["type"])
        tsql.append(f"    [{column_name}] {column_type},")
    if emit_cloud_events_columns:
        tsql.append("    [__type] VARCHAR(MAX) NOT NULL,")
        tsql.append("    [__source] VARCHAR(MAX) NOT NULL,")
        tsql.append("    [__id] VARCHAR(MAX) NOT NULL,")
        tsql.append("    [__time] DATETIME NULL,")
        tsql.append("    [__subject] VARCHAR(MAX) NULL,")
    tsql.append(");")
    tsql.append("")

    # Add the doc string as table metadata
    if "doc" in schema:
        doc_string = schema["doc"]
        tsql.append(f"EXEC sys.sp_addextendedproperty @name = N'MS_Description', @value = N'{doc_string}', @level0type = N'SCHEMA', @level0name = N'dbo', @level1type = N'TABLE', @level1name = N'{table_name}';")
        tsql.append("")

    doc_string_statement = []
    for field in fields:
        column_name = field["name"]
        if "doc" in field:
            doc = field["doc"]
            doc_string_statement.append(f"EXEC sys.sp_addextendedproperty @name = N'MS_Description', @value = N'{doc}', @level0type = N'SCHEMA', @level0name = N'dbo', @level1type = N'TABLE', @level1name = N'{table_name}', @level2type = N'COLUMN', @level2name = N'{column_name}';")
        # determine if the column type is complex, if it is, add a 'MS_Avro_Schema' extended property
        column_type = field["type"]
        # if the column_type is a union with null, then it is a nullable field, just get the type
        if isinstance(column_type, list) and len(column_type) == 2 and "null" in column_type:
            column_type = [x for x in column_type if x != "null"][0]        
        if "type" in column_type and column_type["type"] in ["array", "map", "record", "enum", "fixed"]:
            doc_string_statement.append(f"EXEC sys.sp_addextendedproperty @name = N'MS_Avro_Schema', @value = N'{json.dumps(field)}', @level0type = N'SCHEMA', @level0name = N'dbo', @level1type = N'TABLE', @level1name = N'{table_name}', @level2type = N'COLUMN', @level2name = N'{column_name}';")
        elif isinstance(schema_list, list):
            # if the column type refers to a complex type, it may in the schema list
            column_schema = next((x for x in schema_list if x["name"] == column_type), None)
            if column_schema:
                doc_string_statement.append(f"EXEC sys.sp_addextendedproperty @name = N'MS_Avro_Schema', @value = N'{json.dumps(column_schema)}', @level0type = N'SCHEMA', @level0name = N'dbo', @level1type = N'TABLE', @level1name = N'{table_name}', @level2type = N'COLUMN', @level2name = N'{column_name}';")


    if doc_string_statement and emit_cloud_events_columns:
        doc_string_statement.extend([
            f"EXEC sys.sp_addextendedproperty @name = N'MS_Description', @value = N'Event type', @level0type = N'SCHEMA', @level0name = N'dbo', @level1type = N'TABLE', @level1name = N'{table_name}', @level2type = N'COLUMN', @level2name = N'__type';",
            f"EXEC sys.sp_addextendedproperty @name = N'MS_Description', @value = N'Context origin/source of the event', @level0type = N'SCHEMA', @level0name = N'dbo', @level1type = N'TABLE', @level1name = N'{table_name}', @level2type = N'COLUMN', @level2name = N'__source';",
            f"EXEC sys.sp_addextendedproperty @name = N'MS_Description', @value = N'Event identifier', @level0type = N'SCHEMA', @level0name = N'dbo', @level1type = N'TABLE', @level1name = N'{table_name}', @level2type = N'COLUMN', @level2name = N'__id';",
            f"EXEC sys.sp_addextendedproperty @name = N'MS_Description', @value = N'Event generation time', @level0type = N'SCHEMA', @level0name = N'dbo', @level1type = N'TABLE', @level1name = N'{table_name}', @level2type = N'COLUMN', @level2name = N'__time';",
            f"EXEC sys.sp_addextendedproperty @name = N'MS_Description', @value = N'Context subject of the event', @level0type = N'SCHEMA', @level0name = N'dbo', @level1type = N'TABLE', @level1name = N'{table_name}', @level2type = N'COLUMN', @level2name = N'__subject';"
        ])
    if doc_string_statement:
        tsql.extend(doc_string_statement)

    with open(tsql_file_path, "w") as tsql_file:
        tsql_file.write("\n".join(tsql))

def map_scalar_type(type, nullable=True):
    if type == "null":
        return "NVARCHAR(MAX)" + ("" if nullable else " NOT") + " NULL"
    elif type == "int":
        return "INT" + ("" if nullable else " NOT") + " NULL"
    elif type == "long":
        return "BIGINT" + ("" if nullable else " NOT") + " NULL"
    elif type == "float":
        return "REAL" + ("" if nullable else " NOT") + " NULL"
    elif type == "double":
        return "FLOAT" + ("" if nullable else " NOT") + " NULL"
    elif type == "boolean":
        return "BIT" + ("" if nullable else " NOT") + " NULL"
    elif type == "bytes":
        return "NVARCHAR(MAX)" + ("" if nullable else " NOT") + " NULL"
    elif type == "string":
        return "NVARCHAR(MAX)" + ("" if nullable else " NOT") + " NULL"
    else:
        return "NVARCHAR(MAX)" + ("" if nullable else " NOT") + " NULL"

def convert_avro_type_to_tsql_type(avroType, nullable=False):
    if isinstance(avroType, list):
        # If the type is an array, then it is a union type. Look whether it's a pair of a scalar type and null:
        itemCount = len(avroType)
        if itemCount > 2:
            return " NVARCHAR(MAX)" + ("" if nullable else " NOT") + " NULL"
        if itemCount == 1:
            return convert_avro_type_to_tsql_type(avroType[0], False)
        else:
            first = avroType[0]
            second = avroType[1]
            if isinstance(first, str) and first == "null":
                return convert_avro_type_to_tsql_type(second, True)
            elif isinstance(second, str) and second == "null":
                return convert_avro_type_to_tsql_type(first, True)
            else: 
                return "NVARCHAR(MAX)" + ("" if nullable else " NOT") + " NULL"
    elif isinstance(avroType, dict):
        type = avroType.get("type")
        nullable = avroType.get("nullable", True)
        if type == "enum":
            return "NVARCHAR(MAX)" + ("" if nullable else " NOT") + " NULL"
        elif type == "fixed":
            return "NVARCHAR(MAX)" + ("" if nullable else " NOT") + " NULL"
        elif type == "string":
            logicalType = avroType.get("logicalType")
            if logicalType == "uuid":
                return "UNIQUEIDENTIFIER" + ("" if nullable else " NOT") + " NULL"
            return "NVARCHAR(MAX)" + ("" if nullable else " NOT") + " NULL"
        elif type == "bytes":
            logicalType = avroType.get("logicalType")
            if logicalType == "decimal":
                return "DECIMAL" + ("" if nullable else " NOT") + " NULL"
            return "NVARCHAR(MAX)" + ("" if nullable else " NOT") + " NULL"
        elif type == "long":
            logicalType = avroType.get("logicalType")
            if logicalType in ["timestamp-millis", "timestamp-micros"]:
                return "DATETIME" + ("" if nullable else " NOT") + " NULL"
            if logicalType in ["time-millis", "time-micros"]:
                return "TIME" + ("" if nullable else " NOT") + " NULL"
            return "BIGINT" + ("" if nullable else " NOT") + " NULL"
        elif type == "int":
            logicalType = avroType.get("logicalType")
            if logicalType == "date":
                return "DATE" + ("" if nullable else " NOT") + " NULL"
            return "INT" + ("" if nullable else " NOT") + " NULL"
        else:
            return map_scalar_type(type) + ("" if nullable else " NOT") + " NULL"
    elif isinstance(avroType, str):
        return map_scalar_type(avroType)  + ("" if nullable else " NOT") + " NULL"
    return "NVARCHAR(MAX)" + ("" if nullable else " NOT") + " NULL"

def map_scalar_type(type):
    if type == "null":
        return "NVARCHAR(MAX)"
    elif type == "int":
        return "INT"
    elif type == "long":
        return "BIGINT"
    elif type == "float":
        return "REAL"
    elif type == "double":
        return "FLOAT"
    elif type == "boolean":
        return "BIT"
    elif type == "bytes":
        return "NVARCHAR(MAX)"
    elif type == "string":
        return "NVARCHAR(MAX)"
    else:
        return "NVARCHAR(MAX)"



import json
import sys
import pyarrow as pa
import pyarrow.parquet as pq

def convert_avro_to_parquet(avro_schema_path, avro_record_type, parquet_file_path, emit_cloud_events_columns=False):
    schema_file = avro_schema_path
    if not schema_file:
        print("Please specify the avro schema file")
        sys.exit(1)
    with open(schema_file, "r") as f:
        schema_json = f.read()

    # Parse the schema as a JSON object
    schema = json.loads(schema_json)

    if isinstance(schema, list) and avro_record_type:
        schema = next((x for x in schema if x["name"] == avro_record_type or x["namespace"]+"."+x["name"] == avro_record_type), None)
        if schema is None:
            print(f"No top-level record type {avro_record_type} found in the Avro schema")
            sys.exit(1)
    elif not isinstance(schema, dict):
        print("Expected a single Avro schema as a JSON object, or a list of schema records")
        sys.exit(1)

    # Get the name and fields of the top-level record
    table_name = schema["name"]
    fields = schema["fields"]

    # Create a list to store the parquet schema
    parquet_schema = []

    # Append the parquet schema with the column names and types
    for field in fields:
        column_name = field["name"]
        column_type = convert_avro_type_to_parquet_type(field["type"])
        parquet_schema.append((column_name, column_type))

    if emit_cloud_events_columns:
        parquet_schema.extend([
            ("__type", pa.string()),
            ("__source", pa.string()),
            ("__id", pa.string()),
            ("__time", pa.timestamp('ns')),
            ("__subject", pa.string())
        ])

    # Create an empty table with the schema
    table = pa.Table.from_batches([], schema=pa.schema(parquet_schema))
    pq.write_table(table, parquet_file_path)

def convert_avro_type_to_parquet_type(avroType):
    if isinstance(avroType, list):
        # If the type is an array, then it is a union type. Look whether it's a pair of a scalar type and null:
        itemCount = len(avroType)
        if itemCount == 1:
            return convert_avro_type_to_parquet_type(avroType[0])
        elif itemCount == 2:
            first = avroType[0]
            second = avroType[1]
            if isinstance(first, str) and first == "null":
                return convert_avro_type_to_parquet_type(second)
            elif isinstance(second, str) and second == "null":
                return convert_avro_type_to_parquet_type(first)
            else:
                return pa.struct( [pa.field(f'{x}Value' if isinstance(x,str) else f'{x.get("name")}Value' if isinstance(x,dict) else f'_{i}', 
                                    convert_avro_type_to_parquet_type(x)) for i, x in enumerate(avroType)])
        else:
            return pa.struct( [pa.field(f'{x}Value' if isinstance(x,str) else f'{x.get("name")}Value' if isinstance(x,dict) else f'_{i}', 
                                    convert_avro_type_to_parquet_type(x)) for i, x in enumerate(avroType)])
    elif isinstance(avroType, dict):
        type = avroType.get("type")
        if type == "array":
            return pa.list_(convert_avro_type_to_parquet_type(avroType.get("items")))
        elif type == "map":
            return pa.map_(pa.string(), convert_avro_type_to_parquet_type(avroType.get("values")))
        elif type == "record":
            fields = avroType.get("fields")
            return pa.struct({field.get("name"): convert_avro_type_to_parquet_type(field.get("type")) for field in fields})
        if type == "enum":
            return pa.string()
        elif type == "fixed":
            return pa.string()
        elif type == "string":
            logicalType = avroType.get("logicalType")
            if logicalType == "uuid":
                return pa.string()
            return pa.string()
        elif type == "bytes":
            logicalType = avroType.get("logicalType")
            if logicalType == "decimal":
                return pa.decimal128(38, 18)
            return pa.string()
        elif type == "long":
            logicalType = avroType.get("logicalType")
            if logicalType in ["timestamp-millis", "timestamp-micros"]:
                return pa.timestamp('ns')
            if logicalType in ["time-millis", "time-micros"]:
                return pa.time64('ns')
            return pa.int64()
        elif type == "int":
            logicalType = avroType.get("logicalType")
            if logicalType == "date":
                return pa.date32()
            return pa.int32()
        else:
            return map_scalar_type(type)
    elif isinstance(avroType, str):
        return map_scalar_type(avroType)
    
    return pa.string()

def map_scalar_type(type):
    if type == "null":
        return pa.string()
    elif type == "int":
        return pa.int32()
    elif type == "long":
        return pa.int64()
    elif type == "float":
        return pa.float32()
    elif type == "double":
        return pa.float64()
    elif type == "boolean":
        return pa.bool_()
    elif type == "bytes":
        return pa.string()
    elif type == "string":
        return pa.string()
    else:
        return pa.string()
